import json
from dataclasses import dataclass
from typing import Dict, Any

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType
from schemey import Schema
from starlette.exceptions import HTTPException
from starlette.requests import Request

from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC
from servey.action.util import with_isolated_references


@dataclass
class BodyParser(RequestParserABC):
    """
    Parser which pulls kwargs from a request body
    """

    schema: Schema
    marshaller: MarshallerABC

    async def parse(self, request: Request) -> Dict[str, Any]:
        body = await request.body()
        params: ExternalItemType = json.loads(body) if body else {}
        error = next(self.schema.iter_errors(params), None)
        if error:
            raise HTTPException(422, str(error))
        kwargs = self.marshaller.load(params)
        return kwargs

    def to_openapi_schema(
        self, path_method: ExternalItemType, components: ExternalItemType
    ):
        schema = with_isolated_references(
            self.schema.schema, self.schema.schema, components
        )
        path_method["requestBody"] = {
            "content": {"application/json": {"schema": schema}},
            "required": True,
        }
        responses: ExternalItemType = path_method["responses"]
        responses["422"] = {"description": "Validation Error"}



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from typing import Optional, Callable

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType

from servey.servey_aws.event_parser.api_gateway_event_parser import (
    ApiGatewayEventParser,
)
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.result_render.api_gateway_result_render import (
    ApiGatewayResultRender,
)
from servey.servey_aws.result_render.factory.result_render_factory_abc import (
    ResultRenderFactoryABC,
)
from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


@dataclass
class ApiGatewayResultRenderFactory(ResultRenderFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    priority: int = 100

    def create(
        self, fn: Callable, event: ExternalItemType, context, parser: EventParserABC
    ) -> Optional[ResultRenderABC]:
        if not isinstance(parser, ApiGatewayEventParser):
            return
        sig = inspect.signature(fn)
        marshaller = self.marshaller_context.get_marshaller(sig.return_annotation)
        return ApiGatewayResultRender(marshaller)



import json
from dataclasses import dataclass
from typing import Any

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


@dataclass
class ApiGatewayResultRender(ResultRenderABC):
    marshaller: MarshallerABC

    def render(self, result: Any) -> ExternalItemType:
        # TODO: Caching headers here?
        dumped = json.dumps(self.marshaller.dump(result))
        response = {
            "statusCode": 200,
            "headers": {"Content-Type": "application/json"},
            "body": dumped,
        }
        return response



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Callable

from marshy.types import ExternalItemType

from servey.servey_aws.event_parser.event_parser_abc import EventParserABC


class EventParserFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create(
        self, fn: Callable, event: ExternalItemType, context
    ) -> Optional[EventParserABC]:
        """Create an event parser"""


def create_parser_factories() -> Tuple[EventParserFactoryABC, ...]:
    from marshy.factory.impl_marshaller_factory import get_impls

    factories = [f() for f in get_impls(EventParserFactoryABC)]
    factories.sort(key=lambda f: f.priority, reverse=True)
    return tuple(factories)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Tuple, Optional

from schemey import get_default_schema_context, SchemaContext

from servey.action.action import get_schema_for_params
from servey.action.finder.found_action import FoundAction
from servey.servey_aws.event_parser.event_parser import EventParser
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
    EventParserFactoryABC,
)
from servey.servey_aws.event_parser.validating_event_parser import ValidatingEventParser


@dataclass
class ValidatingEventParserFactory(EventParserFactoryABC):
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    priority: int = 50
    skip: bool = False

    def create(
        self, action: FoundAction, factories: Tuple[EventParserFactoryABC, ...]
    ) -> Optional[EventParserABC]:
        if self.skip:
            return
        schema = get_schema_for_params(action.fn, self.schema_context)
        self.skip = True
        try:
            for factory in factories:
                parser = factory.create(action, factories)
                if parser:
                    return ValidatingEventParser(parser, schema)
        finally:
            self.skip = False



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional, List

from servey.action.finder.found_action import FoundAction
from servey.action.trigger.web_trigger import WebTrigger
from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC


class RequestParserFactoryABC(ABC):
    @abstractmethod
    def create(
        self,
        action: FoundAction,
        trigger: WebTrigger,
        parser_factories: List[RequestParserFactoryABC],
    ) -> Optional[RequestParserABC]:
        """Render a response"""


def create_parser_factories() -> List[RequestParserFactoryABC]:
    from marshy.factory.impl_marshaller_factory import get_impls

    factories = [f() for f in get_impls(RequestParserFactoryABC)]
    factories.sort(key=lambda f: f.priority, reverse=True)
    return factories



import inspect
from dataclasses import dataclass, replace
from logging import getLogger
from typing import Optional, Callable, List

from servey.action.finder.found_action import FoundAction
from servey.action.trigger.web_trigger import WebTrigger
from servey.servey_starlette.request_parser.factory.request_parser_factory_abc import (
    RequestParserFactoryABC,
)
from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC
from servey.servey_starlette.request_parser.self_parser import SelfParser
from servey.action.util import (
    strip_injected_from_schema,
    wrap_fn_for_injection,
)

LOGGER = getLogger(__name__)


@dataclass
class SelfParserFactory(RequestParserFactoryABC):
    priority: int = 90
    skip: bool = False

    def create(
        self,
        action: FoundAction,
        trigger: WebTrigger,
        parser_factories: List[RequestParserFactoryABC],
    ) -> Optional[RequestParserABC]:
        if self.skip or not action.owner:
            return
        sig = inspect.signature(action.fn)
        if not sig.parameters:
            LOGGER.warning(f"Missing 'self' parameter for {action.fn.__name__}")
            return
        self_name = next(sig.parameters)
        wrapped_action = strip_injected_from_action(action, self_name, action.owner)
        self.skip = True
        try:
            for f in parser_factories:
                parser = f.create(wrapped_action, trigger, parser_factories)
                if parser:
                    return SelfParser(parser, action.owner, self_name)
        finally:
            self.skip = False


def strip_injected_from_action(
    action: FoundAction, inject_at: str, constructor: Callable
) -> FoundAction:
    """
    Create a version of the action given where the injected argument given is stripped out,
    effectively behaving as if it does not exist
    """
    action = FoundAction(
        fn=wrap_fn_for_injection(action.fn, inject_at, constructor),
        action_meta=replace(
            action.action_meta,
            params_schema=strip_injected_from_schema(
                action.action_meta.params_schema, inject_at
            ),
        ),
    )
    return action



from dataclasses import dataclass
from typing import Dict, Any

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

from servey.servey_aws.event_parser.event_parser_abc import EventParserABC


@dataclass
class AppsyncEventParser(EventParserABC):
    marshaller: MarshallerABC

    def parse(self, event: ExternalItemType, context) -> Dict[str, Any]:
        arguments = event["arguments"] or {}
        parsed = self.marshaller.load(arguments)
        # TODO: Pass in an authentication header here?
        return parsed



import importlib
import json
import logging
import os
from typing import Dict

from marshy.types import ExternalItemType, ExternalType

from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
    create_parser_factories,
)
from servey.servey_aws.result_render.factory.result_render_factory_abc import (
    create_render_factories,
)
from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


def invoke(event: ExternalItemType, context) -> ExternalType:
    _LOGGER.info(json.dumps(dict(lambda_event=event)))
    parser = get_event_parser(event, context)
    render = get_result_render(event, context, parser)
    kwargs = parser.parse(event, context)
    result = _ACTION_FUNCTION(**kwargs)
    response = render.render(result)
    return response


def get_event_parser(event: Dict, context) -> EventParserABC:
    for factory in _EVENT_PARSER_FACTORIES:
        parser = factory.create(_ACTION_FUNCTION, event, context)
        if parser:
            return parser


def get_result_render(event: Dict, context, parser: EventParserABC) -> ResultRenderABC:
    for factory in _RESULT_RENDER_FACTORIES:
        render = factory.create(_ACTION_FUNCTION, event, context, parser)
        if render:
            return render


_ACTION_MODULE = importlib.import_module(os.environ["SERVEY_ACTION_MODULE"])
_ACTION_FUNCTION = getattr(_ACTION_MODULE, os.environ["SERVEY_ACTION_FUNCTION"])
_EVENT_PARSER_FACTORIES = create_parser_factories()
_RESULT_RENDER_FACTORIES = create_render_factories()
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO)



from dataclasses import dataclass
from typing import Dict, Any, Optional

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType
from schemey import Schema

from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC


@dataclass
class EventParser(EventParserABC):
    marshaller: MarshallerABC
    schema: Optional[Schema] = None
    auth_kwarg_name: Optional[str] = None
    authorizer: Optional[AuthorizerABC] = None
    auth_marshaller: Optional[MarshallerABC[Authorization]] = None

    def parse(self, event: ExternalItemType, context) -> Dict[str, Any]:
        auth_kwarg_value = None
        if self.auth_kwarg_name:
            auth_kwarg_value = event.pop(self.auth_kwarg_name, None)
        if self.schema:
            self.schema.validate(event)
        parsed = self.marshaller.load(event)
        if auth_kwarg_value:
            if isinstance(auth_kwarg_value, str):
                parsed[self.auth_kwarg_name] = self.authorizer.authorize(
                    auth_kwarg_value
                )
            elif self.auth_marshaller:
                # Given that we can invoke a lambda directly requires trust, we allow passing the authorization
                # directly. Maybe this should be revoked in favor of only having signed tokens.
                parsed[self.auth_kwarg_name] = self.auth_marshaller.load(
                    auth_kwarg_value
                )
        return parsed



from dataclasses import dataclass
from typing import Optional, Dict, Any

from marshy.types import ExternalItemType
from starlette.requests import Request

from servey.security.access_control.action_access_control_abc import (
    ActionAccessControlABC,
)
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import AuthorizationError
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC
from servey.action.util import inject_value_at


@dataclass
class AuthorizingParser(RequestParserABC):
    """
    Parser which pulls kwargs from a request body
    """

    authorizer: AuthorizerABC
    parser: RequestParserABC
    inject_at: Optional[str] = None
    access_control: ActionAccessControlABC = ALLOW_ALL

    async def parse(self, request: Request) -> Dict[str, Any]:
        authorization = self.get_authorization(request)
        if not self.access_control.is_executable(authorization):
            raise AuthorizationError()
        kwargs = await self.parser.parse(request)
        if self.inject_at:
            inject_value_at(self.inject_at, kwargs, authorization)
        return kwargs

    def get_authorization(self, request: Request):
        token = request.headers.get("Authorization")
        if not token or not token.startswith("Bearer "):
            return
        token = token[7:]
        return self.authorizer.authorize(token)

    def to_openapi_schema(
        self, path_method: ExternalItemType, components: ExternalItemType
    ):
        responses: ExternalItemType = path_method["responses"]
        if not responses.get("403"):
            responses["403"] = {"description": "Unauthorized"}
        if not path_method.get("security"):
            path_method["security"] = [{"OAuth2PasswordBearer": []}]
        if not components.get("securitySchemas"):
            components["securitySchemes"] = {
                "OAuth2PasswordBearer": {
                    "type": "oauth2",
                    "flows": {"password": {"scopes": {}, "tokenUrl": "/login"}},
                }
            }



from datetime import datetime
from io import StringIO

from marshy.types import ExternalItemType
from ruamel.yaml import YAML

from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    GENERATED_HEADER,
)


class KmsKeyConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    servey_kms_resource_yml_file: str = "serverless_kms_resource.yml"

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.servey_kms_resource_yml_file,
        )
        kms_resource_functions_yml = self.build_kms_resource_functions_yml()
        with open(self.servey_kms_resource_yml_file, "w") as writer:
            writer.write("# ")
            writer.write(GENERATED_HEADER.replace("\n", "\n# "))
            writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
            yaml = YAML()
            s = StringIO()
            yaml.dump(kms_resource_functions_yml, s)
            s = s.getvalue().replace('AWS_ACCOUNT_ID', '!Sub arn:aws:iam::${aws:accountId}:root')
            s = s.replace('SERVEY_KMS_KEY', '!Ref serveyKmsKey')
            writer.write(s)

    def build_kms_resource_functions_yml(self) -> ExternalItemType:
        kms_resource = {
            "Resources": {
                "serveyKmsKey": {
                    "Type": "AWS::KMS::Key",
                    "Properties": {
                        "Description": "Key used for servey authentication and authorization",
                        "Enabled": True,
                        "KeySpec": "RSA_4096",
                        "KeyUsage": "SIGN_VERIFY",
                        "KeyPolicy": {
                            "Version": "2012-10-17",
                            "Id": "serveyKmsKey",
                            "Statement": {
                                "Sid": "Allow administration of the key",
                                "Effect": "Allow",
                                "Principal": {
                                    "AWS": "AWS_ACCOUNT_ID",
                                },
                                "Action": "kms:*",
                                "Resource": "*",
                            },
                        },
                    },
                },
                "serveyKmsKeyAlias": {
                    "Type": 'AWS::KMS::Alias',
                    "Properties": {
                        "AliasName": "alias/serveyKmsKey",
                        "TargetKeyId": "SERVEY_KMS_KEY",
                    },
                },
            },
        }
        return kms_resource



from dataclasses import dataclass
from typing import Callable

from marshy.types import ExternalItemType

from servey.action.action_meta import ActionMeta
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


@dataclass
class LambdaHandler:
    fn: Callable
    action_meta: ActionMeta
    event_parser: EventParserABC
    result_render: ResultRenderABC

    def __call__(self, event: ExternalItemType, context):
        kwargs = self.event_parser.parse(event, context)
        result = self.fn(**kwargs)
        rendered = self.result_render.render_result(result)
        return rendered



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Callable

from marshy.types import ExternalItemType

from servey.action.finder.found_action import FoundAction
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


class ResultRenderFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create(
        self, fn: Callable, event: ExternalItemType, context, parser: EventParserABC
    ) -> Optional[ResultRenderABC]:
        """Create an event parser"""


def create_render_factories() -> Tuple[ResultRenderFactoryABC, ...]:
    from marshy.factory.impl_marshaller_factory import get_impls

    factories = [f() for f in get_impls(ResultRenderFactoryABC)]
    factories.sort(key=lambda f: f.priority, reverse=True)
    return tuple(factories)



from dataclasses import dataclass
from typing import Any, Optional, Dict

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType
from schemey import Schema
from starlette.responses import JSONResponse, Response

from servey.action.util import with_isolated_references
from servey.servey_starlette.response_render.response_render_abc import (
    ResponseRenderABC,
)


@dataclass
class BodyRender(ResponseRenderABC):
    marshaller: MarshallerABC
    schema: Optional[Schema] = None

    def render(self, kwargs: Dict[str, Any], result: Any) -> Response:
        dumped = self.marshaller.dump(result)
        if self.schema:
            self.schema.validate(dumped)
        return JSONResponse(dumped)

    def to_openapi_schema(
        self, responses: ExternalItemType, components: ExternalItemType
    ):
        if not self.schema:
            return
        schema = with_isolated_references(
            self.schema.schema, self.schema.schema, components
        )
        responses["200"] = {
            "description": "Successful Response",
            "content": {"application/json": {"schema": schema}},
        }



import json
from dataclasses import dataclass
from typing import Dict, Any, Optional

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC


@dataclass
class ApiGatewayEventParser(EventParserABC):
    marshaller: MarshallerABC
    auth_kwarg_name: Optional[str] = None
    authorizer: Optional[AuthorizerABC] = None

    def parse(self, event: ExternalItemType, context) -> Dict[str, Any]:
        if event.get("httpMethod") in ("POST", "PATCH", "PUT"):
            params = json.loads(event.get("body") or "{}")
        else:
            params = event.get("queryStringParameters") or {}
        parsed = self.marshaller.load(params)
        if self.auth_kwarg_name:
            parsed[self.auth_kwarg_name] = self.get_authorization(event)
        return parsed

    def get_authorization(self, event: ExternalItemType) -> Optional[Authorization]:
        headers = event.get("headers") or {}
        token = (headers.get("Authorization") or "").strip()
        if not token or not token.startswith("Bearer "):
            return
        token = token[7:]
        return self.authorizer.authorize(token)



from dataclasses import dataclass
from typing import Tuple

from schemey import Schema

from servey.action.trigger.trigger_abc import TriggerABC
from servey.security.access_control.action_access_control_abc import (
    ActionAccessControlABC,
)


@dataclass
class ActionMeta:
    name: str
    description: str
    params_schema: Schema
    result_schema: Schema
    access_control: ActionAccessControlABC
    triggers: Tuple[TriggerABC, ...]
    timeout: int = 15



import os
from datetime import datetime

from marshy.types import ExternalItemType
from ruamel.yaml import YAML

from servey.action.finder.action_finder_abc import find_actions_with_trigger_type
from servey.action.trigger.web_trigger import WebTrigger, WebTriggerMethod
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    GENERATED_HEADER,
)


class AppsyncConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    servey_actions_yml_file: str = "serverless_servey_appsync.yml"
    servey_appsync_schema_file: str = "servey_appsync.graphql"

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["custom", "appSync"],
            self.servey_actions_yml_file,
        )
        self.build_appsync_schema()
        servey_action_functions_yml = self.build_servey_action_functions_yml()
        with open(self.servey_actions_yml_file, "w") as writer:
            writer.write("# ")
            writer.write(GENERATED_HEADER.replace("\n", "\n# "))
            writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
            yaml = YAML()
            yaml.dump(servey_action_functions_yml, writer)

    def build_appsync_schema(self):
        from servey.servey_strawberry.schema_factory import new_schema_for_actions

        with open(self.servey_appsync_schema_file, "w") as writer:
            writer.write("# ")
            writer.write(GENERATED_HEADER.replace("\n", "\n# "))
            writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
            schema = str(new_schema_for_actions())
            schema = schema.replace(
                '\n"""Date with time (isoformat)"""\nscalar DateTime\n', ""
            )
            schema = schema.replace("DateTime", "AWSDateTime")
            writer.write(schema)

    def build_servey_action_functions_yml(self) -> ExternalItemType:
        appsync_definitions = {
            "name": os.environ.get("SERVICE_NAME") or "servey_app",
            "authenticationType": "API_KEY",
            "defaultMappingTemplates": {"request": False, "response": False},
            "mappingTemplates": [],
            "dataSources": [],
            "schema": self.servey_appsync_schema_file,
        }
        for action, trigger in find_actions_with_trigger_type(WebTrigger):
            action_meta = action.action_meta
            field = action_meta.name.title().replace("_", "")
            mapping_template = {
                "dataSource": action_meta.name,
                "type": "Query"
                if trigger.method == WebTriggerMethod.GET
                else "Mutation",
                "field": field[0].lower() + field[1:],
                # TODO: Configure caching here!
            }
            appsync_definitions["mappingTemplates"].append(mapping_template)
            data_source = {
                "name": action_meta.name,
                "type": "AWS_LAMBDA",
                "config": {
                    "functionName": action_meta.name,
                },
            }
            if action_meta.description:
                data_source["description"] = action_meta.description.strip()
            appsync_definitions["dataSources"].append(data_source)

        return appsync_definitions



from dataclasses import dataclass
from typing import Dict, Any

from marshy.types import ExternalItemType
from schemey import Schema

from servey.servey_aws.event_parser.event_parser_abc import EventParserABC


@dataclass
class ValidatingEventParser(EventParserABC):
    parser: EventParserABC
    schema: Schema

    def parse(self, event: ExternalItemType, context) -> Dict[str, Any]:
        self.schema.validate(event)
        parsed = self.parser.parse(event, context)
        return parsed



from dataclasses import dataclass
from typing import Callable, Optional, Type

from servey.action.action_meta import ActionMeta


@dataclass
class FoundAction:
    action_meta: ActionMeta
    fn: Callable
    owner: Optional[Type] = None



import inspect
from dataclasses import is_dataclass, fields
from typing import Callable, Optional

from marshy.factory.optional_marshaller_factory import get_optional_type

from servey.security.authorization import Authorization


def get_inject_at(fn: Callable) -> Optional[str]:
    sig = inspect.signature(fn)
    for p in sig.parameters.values():
        annotation = get_optional_type(p.annotation) or p.annotation
        if annotation == Authorization:
            return p.name



import inspect
from abc import abstractmethod
from typing import Optional

from marshy import get_default_context

from servey.action.finder.found_action import FoundAction
from servey.servey_starlette.response_render.body_render import BodyRender
from servey.servey_starlette.response_render.factory.response_render_factory_abc import (
    ResponseRenderFactoryABC,
)
from servey.servey_starlette.response_render.response_render_abc import (
    ResponseRenderABC,
)


class BodyRenderFactory(ResponseRenderFactoryABC):
    priority: int = 50

    def create(self, action: FoundAction) -> Optional[ResponseRenderABC]:
        return_type = inspect.signature(action.fn).return_annotation
        marshaller = get_default_context().get_marshaller(return_type)
        return BodyRender(marshaller, action.action_meta.result_schema)



from dataclasses import dataclass, field
from typing import Tuple, Iterator

from servey.action.finder.action_finder_abc import find_actions
from servey.action.finder.found_action import FoundAction
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
    create_parser_factories,
    EventParserFactoryABC,
)
from servey.servey_aws.lambda_handler import LambdaHandler
from servey.servey_aws.result_render.factory.result_render_factory_abc import (
    ResultRenderFactoryABC,
    create_render_factories,
)
from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


@dataclass
class LambdaHandlerFactory:
    parser_factories: Tuple[EventParserFactoryABC, ...] = field(
        default_factory=create_parser_factories
    )
    render_factories: Tuple[ResultRenderFactoryABC, ...] = field(
        default_factory=create_render_factories
    )

    def create_lambda_handler(self, action: FoundAction) -> LambdaHandler:
        parser = self.create_parser(action)
        render = self.create_render(action)
        return LambdaHandler(action.fn, action.action_meta, parser, render)

    def create_parser(self, action: FoundAction) -> EventParserABC:
        for factory in self.parser_factories:
            parser = factory.create(action, self.parser_factories)
            if parser:
                return parser

    def create_render(self, action: FoundAction) -> ResultRenderABC:
        for factory in self.render_factories:
            render = factory.create(action, self.render_factories)
            if render:
                return render


def create_lambda_handlers() -> Iterator[LambdaHandler]:
    factory = LambdaHandlerFactory()
    for action in find_actions():
        lambda_handler = factory.create_lambda_handler(action)
        yield lambda_handler



from abc import ABC, abstractmethod
from typing import Optional, List

from servey.action.finder.found_action import FoundAction
from servey.servey_starlette.response_render.response_render_abc import (
    ResponseRenderABC,
)


class ResponseRenderFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create(self, action: FoundAction) -> Optional[ResponseRenderABC]:
        """Render a response"""


def create_render_factories() -> List[ResponseRenderFactoryABC]:
    from marshy.factory.impl_marshaller_factory import get_impls

    factories = [f() for f in get_impls(ResponseRenderFactoryABC)]
    factories.sort(key=lambda f: f.priority, reverse=True)
    return factories



from marshy.types import ExternalItemType

from servey.action.action_meta import ActionMeta
from servey.action.trigger.fixed_rate_trigger import FixedRateTrigger
from servey.action.trigger.trigger_abc import TriggerABC
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)


UNITS = {"days": 86400, "hours": 3600, "minutes": 60}


class FixedRateTriggerHandler(TriggerHandlerABC):
    def handle_trigger(
        self,
        action_meta: ActionMeta,
        trigger: TriggerABC,
        lambda_definition: ExternalItemType,
    ):
        if not isinstance(trigger, FixedRateTrigger):
            return
        events = lambda_definition.get("events")
        if not events:
            events = lambda_definition["events"] = []
        for unit, seconds in UNITS:
            if not trigger.interval % seconds:
                events["schedule"] = dict(
                    rate=f"rate({trigger.interval / seconds} {unit})"
                )
                return
        events.append(dict(schedule=dict(rate=f"rate({trigger.interval} seconds)")))



import importlib
import inspect
import pkgutil
import os
from typing import Iterator

from servey.action.finder.action_finder_abc import ActionFinderABC
from servey.action.finder.found_action import FoundAction


class ModuleActionFinder(ActionFinderABC):
    """
    Default implementation of action finder which searches for actions in a particular module
    """

    root_module_name: str = os.environ.get("SERVEY_ACTION_PATH") or "servey_actions"

    def find_actions(self) -> Iterator[FoundAction]:
        module = __import__(self.root_module_name)
        yield from _find_actions_in_module(module)


def _find_actions_in_module(module) -> Iterator[FoundAction]:
    for name, value in module.__dict__.items():
        if hasattr(value, "__servey_action_meta__"):
            yield FoundAction(value.__servey_action_meta__, value)
        elif inspect.isclass(value):
            for f_name, f_value in value.__dict__.items():
                if hasattr(f_value, "__servey_action_meta__"):
                    yield FoundAction(f_value.__servey_action_meta__, f_value, value)
    if not hasattr(module, "__path__"):
        return  # Module was not a package...
    paths = []
    paths.extend(module.__path__)
    module_infos = list(pkgutil.walk_packages(paths))
    for module_info in module_infos:
        sub_module_name = module.__name__ + "." + module_info.name
        sub_module = importlib.import_module(sub_module_name)
        yield from _find_actions_in_module(sub_module)



from dataclasses import dataclass
from typing import Dict, Any

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType
from schemey import Schema
from starlette.exceptions import HTTPException
from starlette.requests import Request

from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC
from servey.action.util import with_isolated_references


@dataclass
class QueryStringParser(RequestParserABC):
    """
    Parser which pulls kwargs from a request query string
    """

    schema: Schema
    marshaller: MarshallerABC

    async def parse(self, request: Request) -> Dict[str, Any]:
        params = dict(request.query_params)
        error = next(self.schema.iter_errors(params), None)
        if error:
            raise HTTPException(422, str(error))
        kwargs = self.marshaller.load(params)
        return kwargs

    def to_openapi_schema(
        self, path_method: ExternalItemType, components: ExternalItemType
    ):
        schema = self.schema
        properties = schema.schema["properties"]
        required = set(schema.schema.get("required") or [])
        path_method["parameters"] = [
            {
                "required": k in required,
                "schema": with_isolated_references(schema.schema, v, components),
                "name": k,
                "in": "query",
            }
            for k, v in properties.items()
        ]
        responses: ExternalItemType = path_method["responses"]
        responses["422"] = {"description": "Validation Error"}



from abc import ABC, abstractmethod
from typing import Any

from marshy import ExternalType


class ResultRenderABC(ABC):
    @abstractmethod
    def render(self, result: Any) -> ExternalType:
        """Render a result"""



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, Callable

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from schemey import SchemaContext, get_default_schema_context

from servey.action.action import get_marshaller_for_params
from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_factory_abc import create_authorizer
from servey.servey_aws.event_parser.api_gateway_event_parser import (
    ApiGatewayEventParser,
)
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.event_parser.factory.event_parser_factory import (
    separate_auth_kwarg,
)
from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
    EventParserFactoryABC,
)


@dataclass
class ApiGatewayEventParserFactory(EventParserFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    priority: int = 100

    def create(
        self, fn: Callable, event: ExternalItemType, context
    ) -> Optional[EventParserABC]:
        if "httpMethod" not in event:
            return
        fn, auth_kwarg_name = separate_auth_kwarg(fn)
        authorizer = None
        if auth_kwarg_name:
            authorizer = create_authorizer()
        marshaller = get_marshaller_for_params(fn, self.marshaller_context)
        return ApiGatewayEventParser(marshaller, auth_kwarg_name, authorizer)



"""
App for case where servey is run in local schedule mode. Scheduled events are run using
threads (As opposed to Celery)
"""
import logging

from servey.action.finder.action_finder_abc import find_actions_with_trigger_type
from servey.action.trigger.fixed_rate_trigger import FixedRateTrigger
from servey.servey_thread.fixed_rate_trigger_thread import FixedRateTriggerThread

_LOGGER = logging.getLogger(__name__)
_THREADS = [
    FixedRateTriggerThread(action.fn, trigger)
    for action, trigger in find_actions_with_trigger_type(FixedRateTrigger)
]



from dataclasses import dataclass
from typing import Dict, Any, Callable

from marshy.types import ExternalItemType
from starlette.requests import Request

from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC


@dataclass
class SelfParser(RequestParserABC):
    parser: RequestParserABC
    constructor: Callable
    self_name: str = "self"

    async def parse(self, request: Request) -> Dict[str, Any]:
        kwargs = await self.parser.parse(request)
        self_ = self.constructor()
        kwargs[self.self_name] = self_
        return kwargs

    def to_openapi_schema(
        self, path_method: ExternalItemType, components: ExternalItemType
    ):
        # Simply delegate - self should not be visible externally!
        return self.parser.to_openapi_schema(path_method, components)



from marshy.types import ExternalItemType

from servey.action.action_meta import ActionMeta
from servey.action.trigger.trigger_abc import TriggerABC
from servey.action.trigger.web_trigger import WebTrigger
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)


class WebTriggerHandler(TriggerHandlerABC):
    def handle_trigger(
        self,
        action_meta: ActionMeta,
        trigger: TriggerABC,
        lambda_definition: ExternalItemType,
    ):
        if not isinstance(trigger, WebTrigger):
            return
        events = lambda_definition.get("events")
        if not events:
            events = lambda_definition["events"] = []
        events.append(
            dict(
                http=dict(path=action_meta.name, method=trigger.method.value, cors=True)
            )
        )



import inspect
from dataclasses import dataclass, field
from inspect import Parameter
from typing import Optional, Tuple, Callable

from strawberry.types import Info

from servey.action.action_meta import ActionMeta
from servey.action.trigger.web_trigger import WebTrigger
from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_strawberry.handler_filter.handler_filter_abc import (
    HandlerFilterABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


@dataclass
class AuthorizationHandlerFilter(HandlerFilterABC):
    priority: int = 120
    authorizer: AuthorizerABC = field(default_factory=get_default_authorizer)
    # What to name the authorization parameter if none exists.
    info_kwarg_name: str = "strawberry_info"

    def filter(
        self,
        fn: Callable,
        action_meta: ActionMeta,
        trigger: WebTrigger,
        schema_factory: SchemaFactory,
    ) -> Tuple[Callable, ActionMeta, bool]:
        sig = inspect.signature(fn)

        parameters = []
        authorization_kwarg_name = None
        for param in sig.parameters.values():
            if param.annotation is Authorization:
                authorization_kwarg_name = param.name
            else:
                parameters.append(param)

        if not authorization_kwarg_name:
            return fn, action_meta, True

        parameters.append(
            Parameter(
                name=self.info_kwarg_name, kind=Parameter.KEYWORD_ONLY, annotation=Info
            )
        )

        def resolver(**kwargs):
            info = kwargs.pop(self.info_kwarg_name)
            authorization = self.get_authorization(info)
            kwargs[authorization_kwarg_name] = authorization
            result = fn(**kwargs)
            return result

        sig = sig.replace(parameters=parameters)
        resolver.__signature__ = sig
        return resolver, action_meta, True

    def get_authorization(self, info: Info) -> Optional[Authorization]:
        authorization = info.context.get("authorization")
        if authorization:
            return authorization
        request = info.context.get("request")
        if request:
            token = request.headers.get("Authorization")
            if token and token.lower().startswith("bearer "):
                token = token[7:]
                authorization = self.authorizer.authorize(token)
                info.context["authorization"] = authorization
                return authorization



from servey.security.access_control.action_access_control_abc import (
    ActionAccessControlABC,
)
from servey.security.authorization import Authorization
from servey.util.singleton_abc import SingletonABC


class AllowAll(SingletonABC, ActionAccessControlABC):
    def is_viewable(self, authorization: Authorization) -> bool:
        return True

    def is_executable(self, authorization: Authorization) -> bool:
        return True


ALLOW_ALL = AllowAll()



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, Callable

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType

from servey.action.action import get_marshaller_for_params
from servey.servey_aws.event_parser.api_gateway_event_parser import (
    ApiGatewayEventParser,
)
from servey.servey_aws.event_parser.appsync_event_parser import AppsyncEventParser
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
    EventParserFactoryABC,
)


@dataclass
class AppsyncEventParserFactory(EventParserFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    priority: int = 100

    def create(
        self, fn: Callable, event: ExternalItemType, context
    ) -> Optional[EventParserABC]:
        info = event.get("info") or {}
        variables = info.get("variables")
        if variables is None:
            return
        marshaller = get_marshaller_for_params(fn, self.marshaller_context)
        return AppsyncEventParser(marshaller)



"""
This module sets up a celery app using any default Action with a FixedRateTrigger
to be run from a celery worker.
All such actions are added to the global namespace as celery tasks
Run this with `celery -A servey.servey_celery.celery_app worker --loglevel=INFO`
"""
import logging
import os
from celery import Celery

# Setup app and tasks...
from servey.action.finder.action_finder_abc import find_actions_with_trigger_type
from servey.action.trigger.fixed_rate_trigger import FixedRateTrigger

_CELERY_BROKER = os.environ.get("CELERY_BROKER")
_LOGGER = logging.getLogger(__name__)
_LOGGER.info(f"Starting celery with broker {_CELERY_BROKER}")
app = Celery(broker=_CELERY_BROKER)
for _action, _ in find_actions_with_trigger_type(FixedRateTrigger):
    globals()[_action.action_meta.name] = app.task(_action)


# noinspection PyUnusedLocal
@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    for action, trigger in find_actions_with_trigger_type(FixedRateTrigger):
        task = globals()[action.action_meta.name]
        sender.add_periodic_task(trigger.interval, task.s())



from dataclasses import dataclass
from typing import Dict, Any, Optional

from marshy.types import ExternalItemType

from servey.security.access_control.action_access_control_abc import (
    ActionAccessControlABC,
)
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import AuthorizationError
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.action.util import inject_value_at


@dataclass
class AuthorizingEventParser(EventParserABC):
    parser: EventParserABC
    authorizer: AuthorizerABC
    inject_at: Optional[str] = None
    access_control: ActionAccessControlABC = ALLOW_ALL

    def parse(self, event: ExternalItemType, context) -> Dict[str, Any]:
        authorization = self.get_authorization(event)
        if not self.access_control.is_executable(authorization):
            raise AuthorizationError()
        parsed = self.parser.parse(event, context)
        if self.inject_at:
            inject_value_at(self.inject_at, parsed, authorization)
        return parsed

    def get_authorization(self, event: ExternalItemType):
        token = event.get("token")
        if token:
            authorization = self.authorizer.authorize(token)
            return authorization



from dataclasses import dataclass, field
from typing import Iterator, List

from starlette.routing import Route

from servey.action.finder.action_finder_abc import find_actions_with_trigger_type
from servey.action.finder.found_action import FoundAction
from servey.action.trigger.web_trigger import WebTrigger
from servey.servey_starlette.action_endpoint import ActionEndpoint
from servey.servey_starlette.request_parser.factory.request_parser_factory_abc import (
    create_parser_factories,
    RequestParserFactoryABC,
)
from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC
from servey.servey_starlette.response_render.factory.response_render_factory_abc import (
    create_render_factories,
    ResponseRenderFactoryABC,
)
from servey.servey_starlette.response_render.response_render_abc import (
    ResponseRenderABC,
)
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC


@dataclass
class ActionRouteFactory(RouteFactoryABC):
    """Utility for mounting actions to fastapi_integration."""

    parser_factories: List[RequestParserFactoryABC] = field(
        default_factory=create_parser_factories
    )
    render_factories: List[ResponseRenderFactoryABC] = field(
        default_factory=create_render_factories
    )
    path_pattern: str = "/actions/{action_name}"

    def create_routes(self) -> Iterator[Route]:
        for action, trigger in find_actions_with_trigger_type(WebTrigger):
            route = self.create_route(action, trigger)
            yield route

    def create_route(self, action: FoundAction, trigger: WebTrigger) -> Route:
        action_endpoint = self.create_action_endpoint(action, trigger)
        route = action_endpoint.to_route()
        return route

    def create_action_endpoint(
        self, action: FoundAction, trigger: WebTrigger
    ) -> ActionEndpoint:
        parser = self.create_parser(action, trigger)
        render = self.create_render(action)
        endpoint = ActionEndpoint(
            name=action.action_meta.name,
            path=self.path_pattern.format(action_name=action.action_meta.name),
            fn=action.fn,
            methods=(trigger.method,),
            parser=parser,
            render=render,
            description=action.action_meta.description,
        )
        return endpoint

    def create_parser(
        self, action: FoundAction, trigger: WebTrigger
    ) -> RequestParserABC:
        for parser_factory in self.parser_factories:
            parser = parser_factory.create(action, trigger, self.parser_factories)
            if parser:
                return parser

    def create_render(self, action: FoundAction) -> ResponseRenderABC:
        for render_factory in self.render_factories:
            render = render_factory.create(action)
            if render:
                return render



"""
App for case where servey is run in local schedule mode. Scheduled events are run using
asyncio (As opposed to Celery)
"""
import asyncio
from logging import getLogger

from servey.action.finder.action_finder_abc import find_actions_with_trigger_type
from servey.action.finder.found_action import FoundAction
from servey.action.trigger.fixed_rate_trigger import FixedRateTrigger

LOGGER = getLogger(__name__)


async def fixed_rate_task(action_: FoundAction, trigger_: FixedRateTrigger):
    while True:
        await asyncio.sleep(trigger_.interval)
        loop = asyncio.get_event_loop()
        loop.call_soon(action_.fn)
        action_.fn()
        executor = action.create_executor()
        executor.execute_async()


for action, trigger in find_actions_with_trigger_type(FixedRateTrigger):
    asyncio.create_task(fixed_rate_task(action, trigger))



from abc import ABC, abstractmethod
from typing import Callable, Tuple, List, TYPE_CHECKING

from servey.action.action_meta import ActionMeta
from servey.action.trigger.web_trigger import WebTrigger

if TYPE_CHECKING:
    from servey.servey_strawberry.schema_factory import SchemaFactory


class HandlerFilterABC(ABC):
    """
    Sometimes the default behaviour from strawberry is not quite what we want, so this allows us to customize
    how a strawberry object is converted to / from regular python
    Examples include adding authorization
    """

    priority: int = 100

    @abstractmethod
    def filter(
        self,
        fn: Callable,
        action_meta: ActionMeta,
        trigger: WebTrigger,
        schema_factory: "SchemaFactory",
    ) -> Tuple[Callable, ActionMeta, bool]:
        """Filter the action given. The callable is a function"""


def create_handler_filters() -> Tuple[HandlerFilterABC]:
    from marshy.factory.impl_marshaller_factory import get_impls

    filters: List[HandlerFilterABC] = [f() for f in get_impls(HandlerFilterABC)]
    filters.sort(key=lambda f: f.priority, reverse=True)
    return tuple(filters)



from abc import ABC, abstractmethod
from typing import Dict, Any

from marshy.types import ExternalItemType


class EventParserABC(ABC):
    @abstractmethod
    def parse(self, event: ExternalItemType, context) -> Dict[str, Any]:
        """Parse kwargs from an event"""



from abc import abstractmethod, ABC
from typing import Any, Dict

from marshy.types import ExternalItemType
from requests import Response


class ResponseRenderABC(ABC):
    @abstractmethod
    def render(self, kwargs: Dict[str, Any], result: Any) -> Response:
        """Render a response"""

    @abstractmethod
    def to_openapi_schema(
        self, responses: ExternalItemType, components: ExternalItemType
    ):
        """Add an openapi description of this render to the openapi path/method/responses given"""



from dataclasses import field, dataclass
from typing import List, Optional

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext

from servey.action.action import get_marshaller_for_params
from servey.action.finder.found_action import FoundAction
from servey.action.trigger.web_trigger import WebTrigger, BODY_METHODS
from servey.servey_starlette.request_parser.body_parser import BodyParser
from servey.servey_starlette.request_parser.factory.request_parser_factory_abc import (
    RequestParserFactoryABC,
)
from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC


@dataclass
class BodyParserFactory(RequestParserFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    priority: int = 50

    def create(
        self,
        action: FoundAction,
        trigger: WebTrigger,
        parser_factories: List[RequestParserFactoryABC],
    ) -> Optional[RequestParserABC]:
        if trigger.method in BODY_METHODS:
            return BodyParser(
                schema=action.action_meta.params_schema,
                marshaller=get_marshaller_for_params(
                    action.fn, self.marshaller_context
                ),
            )



from dataclasses import dataclass
from typing import Any

from marshy import ExternalType
from schemey import Schema

from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


@dataclass
class ValidatingResultRender(ResultRenderABC):
    result_render: ResultRenderABC
    schema: Schema

    def render_result(self, result: Any) -> ExternalType:
        result = self.result_render.render_result(result)
        self.schema.validate(result)
        return result



import inspect
from copy import deepcopy
from typing import Callable

from marshy.types import ExternalItemType, ExternalType
from schemey import Schema


def with_isolated_references(
    root: ExternalItemType, schema: ExternalType, components: ExternalItemType
):
    if isinstance(schema, dict):
        ref = schema.get("$ref")
        if ref is not None:
            ref = ref.split("/")
            referenced_schema = root
            for r in ref:
                referenced_schema = referenced_schema[r]
            name = referenced_schema.get("name")
            if name not in components:
                components[name] = schema  # Prevent infinite loop
                components[name] = with_isolated_references(
                    root, referenced_schema, components
                )
            return {"$ref": f"components/{name}"}
        schema = {
            k: with_isolated_references(root, v, components) for k, v in schema.items()
        }
        return schema
    elif isinstance(schema, list):
        schema = [with_isolated_references(root, v, components) for v in schema]
        return schema
    else:
        return schema


def inject_value_at(path: str, current, value):
    path = path.split(".")
    parent_path = path[:-1]
    for p in parent_path:
        if hasattr(current, p):
            current = getattr(current, p)
        else:
            current = current[p]
    p = path[-1]
    if hasattr(current, p):
        setattr(current, p, value)
    else:
        current[p] = value


# noinspection PyTypeChecker
def strip_injected_from_schema(schema: Schema, inject_at: str) -> Schema:
    result = deepcopy(schema.schema)
    current = result
    path = inject_at.split(".")
    for p in path[:-1]:
        current = current["properties"][p]
    del current["properties"][path[-1]]
    if current["required"]:
        current["required"] = [r for r in current["required"] if r != path[-1]]
    result_schema = Schema(current, schema.python_type)
    return result_schema


# noinspection PyTypeChecker
def wrap_fn_for_injection(
    fn: Callable, inject_at: str, value_factory: Callable
) -> Callable:
    sig = inspect.signature(fn)
    path = inject_at.split(".")
    if len(path) == 1:
        sig = sig.replace(
            parameters=tuple(p for p in sig.parameters.values() if p.name != inject_at)
        )

    def wrapper(**kwargs):
        value = value_factory()
        inject_value_at(inject_at, kwargs, value)
        result = fn(**kwargs)
        return result

    wrapper.__signature__ = sig
    return wrapper



from datetime import datetime

from marshy.factory.impl_marshaller_factory import get_impls
from marshy.types import ExternalItemType
from ruamel.yaml import YAML

from servey.action.finder.action_finder_abc import find_actions
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    GENERATED_HEADER,
)


class ActionFunctionConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    servey_actions_yml_file: str = "serverless_servey_actions.yml"

    def configure(self, main_serverless_yml_file: str):
        ensure_ref_in_file(
            main_serverless_yml_file, ["functions"], self.servey_actions_yml_file
        )
        servey_action_functions_yml = self.build_servey_action_functions_yml()
        with open(self.servey_actions_yml_file, "w") as writer:
            writer.write("# ")
            writer.write(GENERATED_HEADER.replace("\n", "\n# "))
            writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
            yaml = YAML()
            yaml.dump(servey_action_functions_yml, writer)

    @staticmethod
    def build_servey_action_functions_yml() -> ExternalItemType:
        lambda_definitions = {}
        trigger_handlers = [h() for h in get_impls(TriggerHandlerABC)]
        for action in find_actions():
            action_meta = action.action_meta
            # noinspection PyUnresolvedReferences
            lambda_definition = lambda_definitions[action_meta.name] = dict(
                handler=f"servey.servey_aws.lambda_invoker.invoke",
                timeout=action_meta.timeout,
                environment=dict(
                    SERVEY_ACTION_MODULE=action.fn.__module__,
                    SERVEY_ACTION_FUNCTION=action.fn.__name__,
                ),
            )
            if action_meta.description:
                lambda_definition["description"] = action_meta.description.strip()
            # TODO: Configure caching
            for trigger in action_meta.triggers:
                for handler in trigger_handlers:
                    handler.handle_trigger(action_meta, trigger, lambda_definition)
        return lambda_definitions



import json
from dataclasses import dataclass
from typing import Any

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC

from servey.servey_aws.result_render.result_render_abc import ResultRenderABC


@dataclass
class ResultRender(ResultRenderABC):
    marshaller: MarshallerABC

    def render(self, result: Any) -> ExternalType:
        dumped = self.marshaller.dump(result)
        return dumped



from abc import ABC
from typing import Optional

from servey.security.authorization import Authorization


class ActionAccessControlABC(ABC):
    def is_viewable(self, authorization: Optional[Authorization]) -> bool:
        """Determine if this action's meta may be viewed using the outhorization given"""

    def is_executable(self, authorization: Optional[Authorization]) -> bool:
        """Determine if this action may be executed using the outhorization given"""



import inspect
from typing import Optional, Callable, Tuple

from marshy import get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller.obj_marshaller import attr_config, ObjMarshaller
from marshy.marshaller_context import MarshallerContext
from schemey import Schema, SchemaContext, get_default_schema_context

from servey.action.action_meta import ActionMeta
from servey.errors import ServeyError
from servey.security.access_control.action_access_control_abc import (
    ActionAccessControlABC,
)
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.action.trigger.trigger_abc import TriggerABC
from servey.action.trigger.web_trigger import WebTrigger, WebTriggerMethod


def action(
    fn: Optional[Callable] = None,
    params_schema: Optional[Schema] = None,
    result_schema: Optional[Schema] = None,
    access_control: ActionAccessControlABC = ALLOW_ALL,
    triggers: Tuple[TriggerABC, ...] = (WebTrigger(WebTriggerMethod.POST),),
    timeout: int = ActionMeta.timeout,
):
    """
    Decorator for actions, which may be a function or a class with a designated method_name
    to act as the action. This decorator doesn't really do anything special aside from
    adding metadata to the object which may be interpreted when the action is mounted.

    When mounting, operations like security checks and validations are performed, as well as
    parameter injection for class based actions.
    """

    def wrapper_(fn_: Callable):
        fn_.__servey_action_meta__ = get_meta_for_fn(fn_.__name__, fn_)
        return fn_

    def get_meta_for_fn(name_: str, fn_: Callable):
        nonlocal params_schema, result_schema
        if not params_schema:
            params_schema = get_schema_for_params(fn_)
        if not result_schema:
            result_schema = get_schema_for_result(fn_)
        return ActionMeta(
            name=name_,
            description=fn_.__doc__,
            params_schema=params_schema,
            result_schema=result_schema,
            access_control=access_control,
            triggers=triggers,
            timeout=timeout,
        )

    return wrapper_ if fn is None else wrapper_(fn)


def get_schema_for_params(
    fn: Callable, schema_context: Optional[SchemaContext] = None
) -> Schema:
    if not schema_context:
        schema_context = get_default_schema_context()
    sig = inspect.signature(fn)
    properties = {}
    required = []
    params = list(sig.parameters.values())
    for i, p in enumerate(params):
        if i == 0 and p.name in ("self", "cls"):
            # The action annotation may be applied to a method instead of a function.
            # Unfortunately, this means that the type is unknown at this point - that's sort of ok because
            # a standard finder won't pick up this function anyway without a service annotation.
            continue
        if p.annotation is inspect.Parameter.empty:
            raise ServeyError(f"missing_param_annotation:{fn.__name__}:{p.name}")
        properties[p.name] = schema_context.schema_from_type(p.annotation).schema
        if p.default == inspect.Parameter.empty:
            required.append(p.name)
    json_schema = {
        "type": "object",
        "properties": properties,
        "additionalProperties": False,
        "required": required,
    }
    schema = Schema(json_schema, dict)
    return schema


def get_schema_for_result(fn: Callable) -> Schema:
    schema_context = get_default_schema_context()
    sig = inspect.signature(fn)
    type_ = sig.return_annotation
    if type_ is inspect.Parameter.empty:
        raise ServeyError(f"missing_return_annotation:{fn.__name__}")
    schema = schema_context.schema_from_type(type_)
    return schema


def get_marshaller_for_params(
    fn: Callable, marshaller_context: Optional[MarshallerContext] = None
) -> MarshallerABC:
    if not marshaller_context:
        marshaller_context = get_default_context()
    sig = inspect.signature(fn)
    attr_configs = []
    params = list(sig.parameters.values())
    for p in params:
        if p.annotation is inspect.Parameter.empty:
            raise TypeError(f"missing_param_annotation:{fn.__name__}({p.name}")
        attr_configs.append(
            attr_config(marshaller_context.get_marshaller(p.annotation), p.name)
        )
    marshaller = ObjMarshaller(dict, tuple(attr_configs))
    return marshaller



from __future__ import annotations
from abc import abstractmethod, ABC
from typing import Any, Dict

from marshy.types import ExternalItemType
from starlette.requests import Request


class RequestParserABC(ABC):
    @abstractmethod
    async def parse(self, request: Request) -> Dict[str, Any]:
        """Parse a request"""

    @abstractmethod
    def to_openapi_schema(
        self, path_method: ExternalItemType, components: ExternalItemType
    ):
        """Add an openapi description of this parser to the openapi path/method given"""



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from typing import Optional, Callable, Tuple

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from schemey import get_default_schema_context, SchemaContext

from servey.action.action import get_marshaller_for_params
from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_factory_abc import create_authorizer
from servey.security.util import get_inject_at
from servey.servey_aws.event_parser.event_parser import EventParser
from servey.servey_aws.event_parser.event_parser_abc import EventParserABC
from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
    EventParserFactoryABC,
)


@dataclass
class EventParserFactory(EventParserFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    validate: bool = True
    allow_unsigned_auth: bool = True
    priority: int = 50

    def create(
        self, fn: Callable, event: ExternalItemType, context
    ) -> Optional[EventParserABC]:
        fn, auth_kwarg_name = separate_auth_kwarg(fn)
        authorizer = None
        auth_marshaller = None
        if auth_kwarg_name:
            authorizer = create_authorizer()
            if self.allow_unsigned_auth:
                auth_marshaller = self.marshaller_context.get_marshaller(Authorization)
        marshaller = get_marshaller_for_params(fn, self.marshaller_context)
        # noinspection PyUnresolvedReferences
        schema = (
            get_schema_for_params(fn, self.schema_context) if self.validate else None
        )
        return EventParser(
            marshaller, schema, auth_kwarg_name, authorizer, auth_marshaller
        )


def separate_auth_kwarg(fn: Callable) -> Tuple[Callable, Optional[str]]:
    attr_name = get_inject_at(fn)
    if not attr_name:
        return fn, attr_name

    def noop(**kwargs):
        # This function exists to transfer signature data. it is never called
        raise ValueError()

    sig = inspect.signature(fn)
    sig = sig.replace(
        parameters=[p for p in sig.parameters.values() if p.name != attr_name]
    )
    noop.__signature__ = sig

    return noop, attr_name



from dataclasses import dataclass, field
from typing import List, Optional


from servey.action.finder.found_action import FoundAction
from servey.action.trigger.web_trigger import WebTrigger
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.security.util import get_inject_at
from servey.servey_starlette.request_parser.authorizing_parser import AuthorizingParser
from servey.servey_starlette.request_parser.factory.request_parser_factory_abc import (
    RequestParserFactoryABC,
)
from servey.servey_starlette.request_parser.factory.self_parser_factory import (
    strip_injected_from_action,
)
from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC


@dataclass
class AuthorizingParserFactory(RequestParserFactoryABC):
    priority: int = 80
    authorizer: AuthorizerABC = field(default_factory=get_default_authorizer)
    skip: bool = False

    def create(
        self,
        action: FoundAction,
        trigger: WebTrigger,
        parser_factories: List[RequestParserFactoryABC],
    ) -> Optional[RequestParserABC]:
        if self.skip:
            return
        inject_at = get_inject_at(action.fn)
        if action.action_meta.access_control == ALLOW_ALL and not inject_at:
            return
        wrapped_action = action
        if inject_at:
            # If the wrapped parser tries to invoke the function we supply no authorization
            wrapped_action = strip_injected_from_action(action, inject_at, lambda: None)

        self.skip = True
        try:
            for factory in parser_factories:
                parser = factory.create(wrapped_action, trigger, parser_factories)
                if parser:
                    return AuthorizingParser(
                        self.authorizer,
                        parser,
                        inject_at,
                        action.action_meta.access_control,
                    )
        finally:
            self.skip = False



import logging

from marshy.factory.impl_marshaller_factory import register_impl
from marshy.marshaller_context import MarshallerContext

from servey.action.marshallers.to_second_datetime_marshaller import (
    ToSecondDatetimeMarshaller,
)
from servey.servey_aws.event_parser.factory.appsync_event_parser_factory import (
    AppsyncEventParserFactory,
)

priority = 100
LOGGER = logging.getLogger(__name__)


def configure(context: MarshallerContext):
    from marshy.factory.dataclass_marshaller_factory import DataclassMarshallerFactory

    context.register_factory(
        DataclassMarshallerFactory(priority=101, exclude_dumped_values=tuple())
    )
    context.register_marshaller(ToSecondDatetimeMarshaller())
    configure_action_finder(context)
    configure_auth(context)
    configure_starlette(context)
    configure_aws(context)
    configure_serverless(context)
    configure_strawberry(context)
    configure_strawberry_starlette(context)


def configure_action_finder(context: MarshallerContext):
    from servey.action.finder.action_finder_abc import ActionFinderABC
    from servey.action.finder.module_action_finder import ModuleActionFinder

    register_impl(ActionFinderABC, ModuleActionFinder, context)


def configure_auth(context: MarshallerContext):
    from servey.security.authorizer.authorizer_factory_abc import AuthorizerFactoryABC
    from servey.security.authorizer.jwt_authorizer_factory import JwtAuthorizerFactory

    register_impl(AuthorizerFactoryABC, JwtAuthorizerFactory, context)


def configure_starlette(context: MarshallerContext):
    try:
        configure_starlette_request_parser(context)
        configure_starlette_response_render(context)
        configure_starlette_route_factory(context)
    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Starlette not installed - Skipping")


def configure_starlette_request_parser(context: MarshallerContext):
    from servey.servey_starlette.request_parser.factory.authorizing_parser_factory import (
        AuthorizingParserFactory,
    )
    from servey.servey_starlette.request_parser.factory.body_parser_factory import (
        BodyParserFactory,
    )
    from servey.servey_starlette.request_parser.factory.query_string_parser_factory import (
        QueryStringParserFactory,
    )
    from servey.servey_starlette.request_parser.factory.request_parser_factory_abc import (
        RequestParserFactoryABC,
    )

    register_impl(RequestParserFactoryABC, AuthorizingParserFactory, context)
    register_impl(RequestParserFactoryABC, BodyParserFactory, context)
    register_impl(RequestParserFactoryABC, QueryStringParserFactory, context)


def configure_starlette_response_render(context: MarshallerContext):
    from servey.servey_starlette.response_render.factory.body_render_factory import (
        BodyRenderFactory,
    )
    from servey.servey_starlette.response_render.factory.response_render_factory_abc import (
        ResponseRenderFactoryABC,
    )

    register_impl(ResponseRenderFactoryABC, BodyRenderFactory, context)


def configure_starlette_route_factory(context: MarshallerContext):
    from servey.servey_starlette.route_factory.action_route_factory import (
        ActionRouteFactory,
    )
    from servey.servey_starlette.route_factory.authenticator_route_factory import (
        AuthenticatorRouteFactory,
    )
    from servey.servey_starlette.route_factory.openapi_route_factory import (
        OpenapiRouteFactory,
    )
    from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC

    register_impl(RouteFactoryABC, ActionRouteFactory, context)
    register_impl(RouteFactoryABC, AuthenticatorRouteFactory, context)
    register_impl(RouteFactoryABC, OpenapiRouteFactory, context)


def configure_strawberry(context: MarshallerContext):
    try:
        from servey.servey_strawberry.handler_filter.handler_filter_abc import (
            HandlerFilterABC,
        )
        from servey.servey_strawberry.handler_filter.authorization_handler_filter import (
            AuthorizationHandlerFilter,
        )
        from servey.servey_strawberry.handler_filter.strawberry_type_handler_filter import (
            StrawberryTypeHandlerFilter,
        )

        from servey.servey_strawberry.entity_factory.entity_factory_abc import (
            EntityFactoryABC,
        )
        from servey.servey_strawberry.entity_factory.dataclass_factory import (
            DataclassFactory,
        )
        from servey.servey_strawberry.entity_factory.enum_factory import (
            EnumFactory,
        )
        from servey.servey_strawberry.entity_factory.forward_ref_factory import (
            ForwardRefFactory,
        )
        from servey.servey_strawberry.entity_factory.generic_factory import (
            GenericFactory,
        )
        from servey.servey_strawberry.entity_factory.no_op_factory import (
            NoOpFactory,
        )

        register_impl(HandlerFilterABC, AuthorizationHandlerFilter, context)
        register_impl(HandlerFilterABC, StrawberryTypeHandlerFilter, context)

        register_impl(EntityFactoryABC, DataclassFactory, context)
        register_impl(EntityFactoryABC, EnumFactory, context)
        register_impl(EntityFactoryABC, ForwardRefFactory, context)
        register_impl(EntityFactoryABC, GenericFactory, context)
        register_impl(EntityFactoryABC, NoOpFactory, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Strawberry Module not found: skipping")
        return


def configure_strawberry_starlette(context: MarshallerContext):
    try:
        from servey.servey_starlette.route_factory.route_factory_abc import (
            RouteFactoryABC,
        )
        from servey.servey_strawberry.strawberry_starlette_route_factory import (
            StrawberryStarletteRouteFactory,
        )

        register_impl(RouteFactoryABC, StrawberryStarletteRouteFactory, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Strawberry Module not found: skipping")
        return


def configure_aws(context: MarshallerContext):
    try:
        from servey.servey_aws.authorizer.kms_authorizer_factory import (
            KmsAuthorizerFactory,
        )
        from servey.security.authorizer.authorizer_factory_abc import (
            AuthorizerFactoryABC,
        )

        register_impl(AuthorizerFactoryABC, KmsAuthorizerFactory, context)

        from servey.servey_aws.event_parser.factory.event_parser_factory import (
            EventParserFactory,
        )
        from servey.servey_aws.event_parser.factory.event_parser_factory_abc import (
            EventParserFactoryABC,
        )
        from servey.servey_aws.event_parser.factory.api_gateway_event_parser_factory import (
            ApiGatewayEventParserFactory,
        )

        register_impl(EventParserFactoryABC, EventParserFactory, context)
        register_impl(EventParserFactoryABC, ApiGatewayEventParserFactory, context)
        register_impl(EventParserFactoryABC, AppsyncEventParserFactory, context)

        from servey.servey_aws.result_render.factory.result_render_factory import (
            ResultRenderFactory,
        )
        from servey.servey_aws.result_render.factory.result_render_factory_abc import (
            ResultRenderFactoryABC,
        )
        from servey.servey_aws.result_render.factory.api_gateway_result_render_factory import (
            ApiGatewayResultRenderFactory,
        )

        register_impl(ResultRenderFactoryABC, ResultRenderFactory, context)
        register_impl(ResultRenderFactoryABC, ApiGatewayResultRenderFactory, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("AWS module not found: skipping")


def configure_serverless(context: MarshallerContext):
    try:
        from servey.servey_aws.serverless.yml_config.yml_config_abc import YmlConfigABC
        from servey.servey_aws.serverless.yml_config.action_function_config import (
            ActionFunctionConfig,
        )
        from servey.servey_aws.serverless.yml_config.appsync_config import AppsyncConfig
        from servey.servey_aws.serverless.yml_config.kms_key_config import KmsKeyConfig

        register_impl(YmlConfigABC, ActionFunctionConfig, context)
        register_impl(YmlConfigABC, AppsyncConfig, context)
        register_impl(YmlConfigABC, KmsKeyConfig, context)

        from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
            TriggerHandlerABC,
        )
        from servey.servey_aws.serverless.trigger_handler.web_trigger_handler import (
            WebTriggerHandler,
        )
        from servey.servey_aws.serverless.trigger_handler.fixed_rate_trigger_handler import (
            FixedRateTriggerHandler,
        )

        register_impl(TriggerHandlerABC, WebTriggerHandler, context)
        register_impl(TriggerHandlerABC, FixedRateTriggerHandler, context)

    except ModuleNotFoundError as e:
        LOGGER.error(e)
        LOGGER.info("Serverless module not found: skipping")



from dataclasses import is_dataclass, fields, dataclass
from typing import Type, Optional

import strawberry

from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


class DataclassFactory(EntityFactoryABC):
    priority: int = 120

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if not is_dataclass(annotation):
            return
        type_ = schema_factory.types.get(annotation.__name__)
        if type_:
            return type_

        params = {"__annotations__": {}}
        for f in fields(annotation):
            params["__annotations__"][f.name] = schema_factory.get_type(f.type)

        """
        # noinspection PyDataclass
        params = {
            '__annotations__': {
                f.name: schema_factory.get_type(f.type) for f in fields(annotation)
            },
        }
        """
        wrap_type = dataclass(type(annotation.__name__, tuple(), params))
        type_ = strawberry.type(wrap_type)
        schema_factory.types[annotation.__name__] = type_
        # noinspection PyDataclass
        for field in fields(type_):
            field.type = schema_factory.get_type(field.type)
        return type_

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if not is_dataclass(annotation):
            return
        name = f"{annotation.__name__}Input"
        input_ = schema_factory.inputs.get(name)
        if input_:
            return input_

        # noinspection PyDataclass
        params = {
            "__annotations__": {
                f.name: schema_factory.get_input(f.type) for f in fields(annotation)
            },
        }
        wrap_type = dataclass(type(name, tuple(), params))
        input_ = strawberry.input(wrap_type)
        schema_factory.inputs[name] = input_
        # noinspection PyDataclass
        for field in fields(input_):
            field.type = schema_factory.get_input(field.type)
        return input_



from abc import ABC, abstractmethod
from typing import List

from marshy.factory.impl_marshaller_factory import get_impls
from ruamel.yaml import YAML

GENERATED_HEADER = """
This file was auto generated by servey. Manual modifications may be lost.
""".strip()


class YmlConfigABC(ABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    @abstractmethod
    def configure(self, main_serverless_yml_file: str):
        """Configure the serverless env"""


def configure(main_serverless_yml_file: str):
    for impl in get_impls(YmlConfigABC):
        impl().configure(main_serverless_yml_file)


def ensure_ref_in_file(
    main_serverless_yml_file: str,
    insertion_point: List[str],
    referenced_serverless_yml_file: str,
):
    yaml = YAML()
    with open(main_serverless_yml_file, "r") as reader:
        root = yaml.load(reader)
        reference = "${file(" + referenced_serverless_yml_file + ")}"
        parent = _follow_path(root, insertion_point[:-1])
        references = parent.get(insertion_point[-1])
        if not references:
            parent[insertion_point[-1]] = [reference]
        elif not isinstance(references, list):
            raise ValueError(
                f"{main_serverless_yml_file} : {'.'.join(insertion_point)} should be a list!"
            )
        elif reference in references:
            return  # Already exists - no action needed
        else:
            references.append(reference)
    with open(main_serverless_yml_file, "w") as writer:
        yaml.dump(root, writer)


def _follow_path(root, path: List[str]):
    node = root
    for key in path:
        child = node.get(key)
        if child is None:
            child = node[key] = {}
        node = child
    return node



from dataclasses import dataclass
from typing import Callable, Tuple, Optional

from marshy.types import ExternalItemType
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route

from servey.servey_starlette.request_parser.request_parser_abc import RequestParserABC
from servey.servey_starlette.response_render.response_render_abc import (
    ResponseRenderABC,
)
from servey.action.trigger.web_trigger import WebTriggerMethod


@dataclass
class ActionEndpoint:
    name: str
    path: str
    fn: Callable
    methods: Tuple[WebTriggerMethod, ...]
    parser: RequestParserABC
    render: ResponseRenderABC
    description: Optional[str] = None

    def to_route(self) -> Route:
        return Route(
            self.path,
            name=self.name,
            endpoint=self.execute,
            methods=[m.value for m in self.methods],
        )

    async def execute(self, request: Request) -> Response:
        kwargs = await self.parser.parse(request)
        result = self.fn(**kwargs)
        response = self.render.render(kwargs, result)
        return response

    def to_openapi_schema(self, schema: ExternalItemType):
        paths: ExternalItemType = schema["paths"]
        components: ExternalItemType = schema["components"]
        path = paths.get(self.path)
        if not path:
            path = paths[self.path] = {}
        for method in self.methods:
            responses = {}
            path_method = path[method.value] = {"responses": responses}
            path_method["operationId"] = self.name
            if self.description:
                path_method["summary"] = self.description
            # Tags?
            self.parser.to_openapi_schema(path_method, components)
            self.render.to_openapi_schema(responses, components)
